"""Handle Open positions."""
import asyncio
from logging import getLogger

from .core import MetaTrader, TradePosition, TradeAction, OrderType
from .order import Order

logger = getLogger(__name__)

class Positions:
    """Get Open Positions.

    Attributes:
        symbol (str): Financial instrument name.
        group (str): The filter for arranging a group of necessary symbols. Optional named parameter.
            If the group is specified, the function returns only positions meeting a specified criteria for a symbol name.
        ticket (int): Position ticket.
        mt5 (MetaTrader): MetaTrader instance.
    """
    mt5: MetaTrader = MetaTrader()
    
    def __init__(self, *, symbol: str = "", group: str = "", ticket: int = 0):
        """Get Open Positions.

        Keyword Args:
            symbol (str): Financial instrument name.
            group (str): The filter for arranging a group of necessary symbols. Optional named parameter. If the group
                is specified, the function returns only positions meeting a specified criteria for a symbol name.
            ticket (int): Position ticket

        """
        self.symbol = symbol
        self.group = group
        self.ticket = ticket

    async def positions_total(self) -> int:
        """Get the number of open positions.
        
        Returns:
            int: Return total number of open positions
        """
        return await self.mt5.positions_total()

    async def positions_get(self):
        """Get open positions with the ability to filter by symbol or ticket.
        
        Returns:
            list[TradePosition]: A list of open trade positions
        """
        positions = await self.mt5.positions_get(group=self.group, symbol=self.symbol, ticket=self.ticket)
        if not positions:
            return []
        return [TradePosition(**pos._asdict()) for pos in positions]

    async def close_all(self) -> int:
        """Close all open positions for the trading account.

        Returns:
            int: Return number of positions closed.
        """
        orders = [Order(action=TradeAction.DEAL, price=pos.price_current, position=pos.ticket,
                        type=OrderType(pos.type).opposite,
                        **pos.get_dict(include={'symbol', 'volume'})) for pos in (await self.positions_get())]

        results = await asyncio.gather(*[order.send() for order in orders], return_exceptions=True)
        amount_closed = len([res for res in results if res.retcode == 10009])
        pos = await self.positions_total()
        if pos > 0:
            logger.warning(f'Failed to close {pos} positions')
        else:
            logger.info('All positions closed')
        return amount_closed



# import asyncio
# from pprint import pprint as pp
# from aiomql import Order, Account, ForexSymbol, RAM, OrderType
#
#
# async def test_send_order():
#     await Account().sign_in()
#     symbol = ForexSymbol(name='Volatility 50 Index')
#     await symbol.init()
#     tick = await symbol.info_tick()
#     pips = 100
#     volume = await symbol.compute_volume(amount=100, pips=pips)
#     sls = symbol.trade_stops_level
#     cls = pips * symbol.pip
#     print(sls, cls)
#     sl = tick.ask - (cls)
#     tp = tick.ask + (cls)
#     sls = symbol.trade_stops_level
#     po = sls * symbol.point
#     pi = po * 10
#     # print(sls, po, pi, sl)
#     print(volume)
#     order = Order(symbol=symbol.name, type=OrderType.BUY, sl=sl, tp=tp, volume=volume, price=tick.ask)
#     res = await order.send()
#     pp(res.dict)

asyncio.run(test_send_order())




import logging
from datetime import datetime

from ...utils import dict_to_string
from ...trader import Trader
from ...core.constants import OrderType, OrderTime

from ...result import Result

logger = logging.getLogger()


class DealTrader(Trader):
    """A base class for placing trades based on the number of pips to target"""

    async def create_order(self, *, order_type: OrderType, pips: float = 0):
        """Using the number of target pips it determines the lot size, stop loss and take profit for the order,
        and updates the order object with the values.

        Args:
            order_type (OrderType): Type of order
            pips (float): Target pips
        """
        volume = await self.ram.get_volume(symbol=self.symbol, pips=pips)
        self.order.volume = volume
        self.order.type = order_type
        await self.set_order_limits(pips=pips)



from math import ceil, log10

from ...symbol import Symbol
from ...core.exceptions import VolumeError


class ForexSymbol(Symbol):
    """Subclass of Symbol for Forex Symbols. Handles the conversion of currency and the computation of stop loss,
     take profit and volume.
    """
    @property
    def pip(self):
        """Returns the pip value of the symbol. This is ten times the point value for forex symbols.

        Returns:
            float: The pip value of the symbol.
        """
        return self.point * 10
    async def compute_volume(self, *, amount: float, pips: float, use_minimum: bool = True) -> float:
        """Compute volume given an amount to risk and target pips. Round the computed volume to the nearest step.

        Args:
            amount (float): Amount to risk. Given in terms of the account currency.
            pips (float): Target pips.

        Keyword Args:
            use_minimum (bool): If True, the minimum volume is returned if the computed volume is less than the minimum volume.

        Returns:
            float: volume

        Raises:
            VolumeError: If the computed volume is less than the minimum volume or greater than the maximum volume.
        """
        if (base := self.currency_profit) != (quote := self.account.currency):
            amount = await self.currency_conversion(amount=amount, base=base, quote=quote)

        volume = amount / (self.pip * self.trade_contract_size * pips)
        step = ceil(abs(log10(self.volume_step)))
        volume = round(volume, step)

        if ((volume < self.volume_min) or (volume > self.volume_max)) and not use_minimum:
            raise VolumeError(f'Incorrect Volume. Computed Volume: {volume}; Symbol Max Volume: {self.volume_max}; '
                             f'Symbol Min Volume: {self.volume_min}')

        return max(volume, self.volume_min)



import asyncio
import logging
from typing import Literal
from dataclasses import dataclass
from zoneinfo import ZoneInfo
from datetime import datetime

from ..traders.simple_deal_trader import DealTrader, Trader
from ... import Symbol
from ...strategy import Strategy
from ...core.constants import TimeFrame, OrderType
from ...candle import Candle, Candles

logger = logging.getLogger(__name__)


@dataclass
class Entry:
    """
    Entry class for FingerTrap strategy.Will be used to store entry conditions and other entry related data.

    Attributes:
        bearish (bool): True if the market is bearish
        bullish (bool): True if the market is bullish
        ranging (bool): True if the market is ranging
        snooze (float): Time to wait before checking for entry conditions
        trend (str): The current trend of the market
        last_candle (Candle): The last candle of the market
        new (bool): True if the last candle is new
        order_type (OrderType): The type of order to place
        pips (int): The number of pips to place the order from the current price
    """
    bearish: bool = False
    bullish: bool = False
    ranging: bool = True
    trending: bool = False
    trend: Literal['ranging', 'bullish', 'bearish'] = 'ranging'
    snooze: float = 0
    last_trend_time: float = 0
    last_entry_time: float = 0
    new: bool = True
    order_type: OrderType | None = None
    pips: float = 10

    def update(self, **kwargs):
        fields = self.__dict__
        for key in kwargs:
            if key in fields:
                setattr(self, key, kwargs[key])
        match self.trend:
            case 'ranging':
                self.ranging = True
                self.trending = self.bullish = self.bearish = False
            case 'bullish':
                self.ranging = self.bearish = False
                self.bullish = self.trending = True
            case 'bearish':
                self.ranging = self.bullish = False
                self.bearish = self.trending = True


class FingerTrap(Strategy):
    trend_time_frame: TimeFrame
    entry_time_frame: TimeFrame
    trend: int
    fast_period: int
    slow_period: int
    entry_period: int
    parameters: dict
    prices: Candles
    name = "FingerTrap"
    pips: float
    entry_interval: float
    entry_candles_count: int
    trend_candles_count: int
    def __init__(self, *, symbol: Symbol, params: dict | None = None, trader: Trader = None):
        super().__init__(symbol=symbol, params=params)
        self.trend = self.parameters.get('trend', 3)
        self.fast_period = self.parameters.setdefault('fast_period', 8)
        self.slow_period = self.parameters.setdefault('slow_period', 34)
        self.entry_time_frame = self.parameters.setdefault('entry_time_frame', TimeFrame.M5)
        self.trend_time_frame = self.parameters.setdefault('trend_time_frame', TimeFrame.H1)
        self.trader = trader or DealTrader(symbol=self.symbol)
        self.pips = self.parameters.setdefault('pips', 10)
        self.entry: Entry = Entry(snooze=self.trend_time_frame.time, pips=self.pips)
        self.entry_period = self.parameters.setdefault('entry_period', 8)
        self.interval = self.parameters.setdefault('interval', 60)

        self.trend_candles_count = self.parameters.setdefault('trend_candles_count',
                                                              86400 // self.trend_time_frame.time)
        self.trend_candles_count = max(self.trend_candles_count, self.slow_period)
        self.entry_candles_count = self.trend_candles_count * (self.trend_time_frame.time // self.entry_time_frame.time)
        self.entry_candles_count = max(self.entry_candles_count, self.entry_period)
        self.tz = ZoneInfo('UTC')

    async def check_trend(self):
        try:
            candles = await self.symbol.copy_rates_from(timeframe=self.trend_time_frame,
                                                        date_from=datetime.now(tz=self.tz), count=self.trend_candles_count)
            current = candles[-1]
            if current.time > self.entry.last_trend_time:
                self.entry.update(new=True, last_trend_time=current.time)
            else:
                self.entry.update(new=False)
                return
            candles.ta.ema(length=self.slow_period, append=True, fillna=0)
            candles.ta.ema(length=self.fast_period, append=True, fillna=0)
            cols = {f'EMA_{self.fast_period}': 'fast', f'EMA_{self.slow_period}': 'slow'}
            candles.rename(inplace=True, **cols)
            # Compute 
            fas = candles.ta_lib.above(candles.fast, candles.slow)
            fbs = candles.ta_lib.below(candles.fast, candles.slow)
            paf = candles.ta_lib.above(candles.close, candles.fast)
            pbf = candles.ta_lib.below(candles.close, candles.slow)
            
            candles.data[fas.name] = fas
            candles.data[fbs.name] = fbs
            candles.data[paf.name] = paf
            candles.data[pbf.name] = pbf
            trend = candles[-self.trend: -1]

            if all(c for c in trend.fast_A_slow) and all(c for c in trend.close_A_fast):
                self.entry.update(trend='bullish')

            elif all(c for c in trend.fast_B_slow) and all(c for c in trend.close_B_slow):
                self.entry.update(trend='bearish')

            else:
                self.entry.update(trend='ranging', snooze=self.trend_time_frame.time)
        except Exception as exe:
            logger.error(f'{exe}. Error in {self.__class__.__name__}.check_trend')

    async def confirm_trend(self):
        try:
            candles = await self.symbol.copy_rates_from(timeframe=self.entry_time_frame, date_from=datetime.now(tz=self.tz),
                                                        count=self.entry_candles_count)
            if candles[-1].time <= self.entry.last_trend_time:
                self.entry.update(trending=False)
                return

            candles.ta.ema(length=self.entry_period, append=True, fillna=0)
            candles.rename(**{f'EMA_{self.entry_period}': 'ema'})
            cxae = candles.ta_lib.cross(candles.close, candles.ema)
            cxbe = candles.ta_lib.cross(candles.close, candles.ema, above=False)
            candles[cxae.name] = cxae
            candles[cxbe.name] = cxbe
            if self.entry.bullish and candles[-1].close_XA_ema:
                self.entry.update(snooze=self.entry_time_frame.time, order_type=OrderType.BUY)
            elif self.entry.bearish and candles[-1].close_XB_ema:
                self.entry.update(snooze=self.entry_time_frame.time, order_type=OrderType.SELL)
            else:
                self.entry.update(snooze=self.interval, order_type=None)
        except Exception as exe:
            logger.error(f'{exe} Error in {self.__class__.__name__}.confirm_trend')

    async def watch_market(self):
        if not self.entry.trending:
            await self.check_trend()
        if self.entry.trending:
            await self.confirm_trend()

    async def trade(self):
        print(f'Trading {self.symbol}')
        while True:
            try:
                await self.watch_market()
                if not self.entry.new:
                    await asyncio.sleep(0.1)
                    continue

                if self.entry.order_type is None:
                    await self.sleep(self.entry.snooze)
                    continue

                await self.trader.place_trade(order_type=self.entry.order_type, pips=self.entry.pips,
                                              params=self.parameters)
                await self.sleep(self.entry.snooze)
            except Exception as err:
                logger.error(f"Error: {err}\t Symbol: {self.symbol} in {self.__class__.__name__}.trade")
                await self.sleep(self.trend_time_frame.time)
                continue



import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Sequence

from .strategy import Strategy
from .symbol import Symbol


class Executor:
    """Executor class for running multiple strategies on multiple symbols concurrently.

    Attributes:
        executor (ThreadPoolExecutor): The executor object.
        workers (list): List of strategies.

    """
    def __init__(self):
        self.executor = ThreadPoolExecutor
        self.workers: list[type(Strategy)] = []

    def add_workers(self, strategies: Sequence[type(Strategy)]):
        """Add multiple strategies at once

        Args:
            strategies (Sequence[Strategy]): A sequence of strategies.
        """
        self.workers.extend(strategies)

    def remove_workers(self, *symbols: Sequence[Symbol]):
        """Removes any worker running on a symbol not successfully initialized.
        
        Args:
            *symbols: Successfully initialized symbols.
        """
        self.workers = [worker for worker in self.workers if worker.symbol in symbols]

    def add_worker(self, strategy: type(Strategy)):
        """Add a strategy instance to the list of workers

        Args:
            strategy (Strategy): A strategy object
        """
        self.workers.append(strategy)

    @staticmethod
    def run(strategy: type(Strategy)):
        """Wraps the coroutine trade method of each strategy with 'asyncio.run'.

        Args:
            strategy (Strategy): A strategy object
        """
        asyncio.run(strategy.trade())

    async def execute(self, workers: int = 0):
        """Run the strategies with a threadpool executor.

        Args:
            workers: Number of workers to use in executor pool. Defaults to zero which uses all workers.

        Notes:
            No matter the number specified, the executor will always use a minimum of 5 workers.
        """
        workers = workers or len(self.workers)
        workers = max(workers, 5)
        loop = asyncio.get_running_loop()
        with self.executor(max_workers=workers) as executor:
            [loop.run_in_executor(executor, self.run, worker) for worker in self.workers]



"""Risk Assessment and Management"""
from .account import Account
from .symbol import Symbol


class RAM:
    account: Account = Account()
    risk_to_reward: float
    risk: float
    amount: float
    pips: float
    volume: float

    def __init__(self, **kwargs):
        """Risk Assessment and Management. All provided keyword arguments are set as attributes.

        Args:
            kwargs (Dict): Keyword arguments.

        Defaults:
            risk_to_reward (float): Risk to reward ratio 1
            risk (float): Percentage of account balance to risk per trade 0.01 # 1%
            amount (float): Amount to risk per trade in terms of account currency 0
            pips (float): Target pips 0
            volume (float): Volume to trade 0


        """
        self.risk_to_reward = kwargs.pop('risk_to_reward', 1)
        self.risk = kwargs.pop('risk', 0.01)
        self.amount = kwargs.pop('amount', 0)
        self.pips = kwargs.pop('pips', 0)
        self.volume = kwargs.pop('volume', 0)
        [setattr(self, key, value) for key, value in kwargs.items()]

    async def get_amount(self, risk: float = 0) -> float:
        """Calculate the amount to risk per trade as a percentage of free margin.

        Keyword Args:
            risk (float): Percentage of account balance to risk per trade. Defaults to zero.

        Returns:
            float: Amount to risk per trade
        """
        await self.account.refresh()
        risk = risk or self.risk
        return self.account.margin_free * risk

    async def get_volume(self, *, symbol: Symbol, pips: float = 0, amount: float = 0) -> float:
        """Calculate the volume to trade. if pips is not provided, the pips attribute is used.
        If the amount attribute or amount argument is zero, the amount is calculated using the get_amount method based on the risk.

        Args:
            symbol (Symbol): Financial instrument

        Keyword Args:
            pips (float): Target pips. Defaults to zero.
            amount (float): Amount to risk per trade. Defaults to zero.

        Returns:
            float: Volume to trade
        """
        pips = pips or self.pips
        amount = amount or self.amount or await self.get_amount()
        return await symbol.compute_volume(amount=amount, pips=pips)




from logging import getLogger
from typing import Type

from .core.models import AccountInfo, SymbolInfo
from .core.exceptions import LoginError

logger = getLogger(__name__)


class Account(AccountInfo):
    """A class for managing a trading account. A singleton class.
    A subclass of AccountInfo. All AccountInfo attributes are available in this class.

    Attributes:
        connected (bool): Status of connection to MetaTrader 5 Terminal
        symbols (set[SymbolInfo]): A set of available symbols for the financial market.

    Notes:
        Other Account properties are defined in the AccountInfo class.
    """
    connected: bool
    symbols = set()
    
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = super().__new__(cls)
        return cls._instance

    async def refresh(self):
        """
        Refreshes the account instance with the latest account details from the MetaTrader 5 terminal
        """
        account_info = await self.mt5.account_info()
        acc = account_info._asdict()
        self.set_attributes(**acc)

    @property
    def account_info(self) -> dict:
        """Get account login, server and password details. If the login attribute of the account instance returns
         a falsy value, the config instance is used to get the account details.

        Returns:
            dict: A dict of login, server and password details

        Note:
            This method will only look for config details in the config instance if the login attribute of the
            account Instance returns a falsy value
        """
        acc_info = self.get_dict(include={'login', 'server', 'password'})
        return acc_info if acc_info['login'] else self.config.account_info()

    async def __aenter__(self) -> 'Account':
        """Connect to a trading account and return the account instance.
        Async context manager for the Account class.

        Returns:
            Account: An instance of the Account class

        Raises:
            LoginError: If login fails
        """
        res = await self.sign_in()
        if not res:
            raise LoginError('Login failed')
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.mt5.shutdown()
        self.connected = False

    async def sign_in(self) -> bool:
        """Connect to a trading account.

        Returns:
            bool: True if login was successful else False
        """
        await self.mt5.initialize(**self.account_info)
        self.connected = await self.mt5.login(**self.account_info)
        if self.connected:
            await self.refresh()
            self.symbols = await self.symbols_get()
            return self.connected
        await self.mt5.shutdown()
        return False

    def has_symbol(self, symbol: str | Type[SymbolInfo]):
        """Checks to see if a symbol is available for a trading account

        Args:
            symbol (str | SymbolInfo):

        Returns:
            bool: True if symbol is present otherwise False
        """
        try:
            symbol = SymbolInfo(name=symbol) if isinstance(symbol, str) else symbol
            return symbol in self.symbols
        except Exception as err:
            logger.warning(f'Error: {err}; {symbol} not available in this market')
            return False

    async def symbols_get(self) -> set[SymbolInfo]:
        """Get all financial instruments from the MetaTrader 5 terminal available for the current account.

        Returns:
            set[Symbol]: A set of available symbols.
        """
        syms = await self.mt5.symbols_get()
        return {SymbolInfo(name=sym.name) for sym in syms}



import asyncio
from typing import Type, Iterable, TypeVar
import logging

from .executor import Executor
from .account import Account
from .symbol import Symbol as _Symbol
from .strategy import Strategy as _Strategy

logger = logging.getLogger(__name__)
Strategy = TypeVar('Strategy', bound=_Strategy)
Symbol = TypeVar('Symbol', bound=_Symbol)


class Bot:
    """The bot class. Create a bot instance to run your strategies.

    Attributes:
        account (Account): Account Object.
        executor: The default thread executor.
        symbols (set[Symbols]): A set of symbols for the trading session
    """
    account: Account = Account()
    executor = Executor()
    symbols = set()

    async def initialize(self):
        """Prepares the bot by signing in to the trading account and initializing the symbols for the trading session.

        Raises:
            SystemExit if sign in was not successful
        """
        init = await self.account.sign_in()
        logger.info("Login Successful")
        if not init:
            logger.warning('Unable to sign in to MetaTrder 5 Terminal')
            raise SystemExit
        await self.init_symbols()
        self.executor.remove_workers(*self.symbols)

    def execute(self):
        """Execute the bot.
        """
        asyncio.run(self.start())

    async def start(self):
        """Starts the bot by calling the initialize method and running the strategies in the executor.
        """
        await self.initialize()
        await self.executor.execute()

    def add_strategy(self, strategy: Strategy):
        """Add a strategy to the executor. An added strategy will only run if it's symbol was successfully initialized.

        Args:
            strategy (Strategy): A Strategy instance to run on bot

        Notes:
            Make sure the symbol has been added to the market
        """
        self.symbols.add(strategy.symbol)
        self.executor.add_worker(strategy)

    def add_strategies(self, strategies: Iterable[Strategy]):
        """Add multiple strategies at the same time

        Args:
            strategies: A list of strategies
        """
        [self.add_strategy(strategy) for strategy in strategies]

    def add_strategy_all(self, *, strategy: Type[Strategy], params: dict | None = None):
        """Use this to run a single strategy on all available instruments in the market using the default parameters
        i.e one set of parameters for all trading symbols

        Keyword Args:
            strategy (Strategy): Strategy class
            params (dict): A dictionary of parameters for the strategy
        """
        [self.add_strategy(strategy(symbol=symbol, params=params)) for symbol in self.symbols]

    async def init_symbols(self):
        """Initialize the symbols for the current trading session. This method is called internally by the bot.
        """
        syms = [self.init_symbol(symbol) for symbol in self.symbols]
        await asyncio.gather(*syms, return_exceptions=True)

    async def init_symbol(self, symbol: Symbol) -> Symbol:
        """Initialize a symbol before the beginning of a trading sessions.
        Removes it from the list of symbols if it was not successfully initialized or not available
        for the current market.

        Args:
            symbol (Symbol): Symbol object to be initialized

        Returns:
            Symbol: if successfully initialized
        """
        if self.account.has_symbol(symbol):
            init = await symbol.init()
            if init:
                return symbol
            self.symbols.discard(symbol)
            logger.warning(f'Unable to initialize symbol {symbol}')

        self.symbols.remove(symbol)
        logger.warning(f'{symbol} not a available for this market')



"""This module contains the Records class, which is used to read and update trade records from csv files."""

import asyncio
from datetime import datetime
from pathlib import Path
import csv

from .history import History
from .core import Config


class Records:
    """This utility class read trade records from csv files, and update them based on their closing positions.

    Attributes:
        config: Config object
        records_dir(Path): Path to directory containing record of placed trades, If not given takes the default
            from the config
    """
    config: Config = Config()
    
    def __init__(self, records_dir: Path = ''):
        """Initialize the Records class.
        Keyword Args:
            records_dir (Path): Path to directory containing record of placed trades.

        """
        self.records_dir = records_dir or self.config.records_dir

    async def get_records(self):
        """Get trade records from records_dir folder

        Yields:
            files: Trade record files
        """
        for file in self.records_dir.iterdir():
            if file.is_file() and file.name.endswith('.csv'):
                yield file

    async def read_update(self, file: Path):
        """Read and update trade records

        Args:
            file: Trade record file
        """
        fr = open(file, mode='r', newline='')
        reader = csv.DictReader(fr)
        rows = [row for row in reader]
        rows = await self.update_rows(rows)
        fr.close()
        fw = open(file, mode='w', newline='')
        writer = csv.DictWriter(fw, fieldnames=reader.fieldnames)
        writer.writeheader()
        writer.writerows(rows)
        fw.close()

    async def update_rows(self, rows: list[dict]) -> list[dict]:
        """Update the rows of entered trades in the csv file with the actual profit.

        Args:
            rows: A list of dictionaries from the dictionary writer object of the csv file.

        Returns:
            list[dict]: A list of dictionaries with the actual profit and win status.
        """
        start = float(min(rows, key=lambda r: r['time'])['time'])
        end = datetime.utcnow().timestamp()
        his = History(date_from=start, date_to=end)
        await his.init(orders=False)
        deals = {str(deal.position_id): deal.profit for deal in his.deals}
        for row in rows:
            if (deal := row['order']) in deals:
                profit = deals[deal]
                row.update(actual_profit=profit, win=profit > 0)
        return rows

    async def update_records(self):
        """Update trade records in the records_dir folder."""
        records = [self.read_update(record) async for record in self.get_records()]
        await asyncio.gather(*records)

    async def update_record(self, file: Path | str):
        """Update a single trade record file."""
        await self.read_update(file)
